//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     rvars.cxx
// Purpose:    The set of variables in a rule.
// Last Change:27.09.2001
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2000-2001 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "out.h"
#include "var.h"
#include "rvar.h"

#include "rvars.h"

//=============================================================================
// Exported Functions:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor:
//-----------------------------------------------------------------------------

rvars_c::rvars_c(void)
{
	// Initialize Magic Number:
#ifdef VER_DEBUG
	Magic = RVARS_MAGIC;
#endif

	// Initialize Member Variables:
	First = 0;
	Last = 0;
	NumVars = 0;
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

rvars_c::~rvars_c(void)
{
#ifdef VER_DEBUG
	Magic = 0;
#endif
}

//-----------------------------------------------------------------------------
// Debugging Support (Integrity Check):
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
bool_t	rvars_c::valid(void) const
{
	int i;
	rvar_t p;

	// Check object pointer:
	CHECK(this != 0, "rvar_c::valid: NIL object");

	// The magic number must be right:
	if(Magic != RVARS_MAGIC)
		return(BOOL_FALSE);

	// The number of variables cannot be negative:
	if(NumVars < 0)
		return(BOOL_FALSE);

	// It can also not be too large:
	if(NumVars >= VER_MAX_VARS)
		return(BOOL_FALSE);

	// If the number of variables is 0, the pointers must be NIL:
	if(NumVars == 0 && First != 0)
		return(BOOL_FALSE);
	if(NumVars == 0 && Last != 0)
		return(BOOL_FALSE);

	// If the number of variables is not 0, the pointers cannot be NIL:
	if(NumVars != 0 && First == 0)
		return(BOOL_FALSE);
	if(NumVars != 0 && Last == 0)
		return(BOOL_FALSE);

	// Check the offsets and the Last pointer:
	i = 0;
	for(p = First; p; p = p->get_next()) {
		if(p->get_offset() != i)
			return(BOOL_FALSE);
		i++;
		if(i == NumVars && Last != p)
			return(BOOL_FALSE);
	}

	// Check the number of variables:
	if(i != NumVars)
		return(BOOL_FALSE);

	// Ok:
	return(BOOL_TRUE);
}
#endif

//-----------------------------------------------------------------------------
// find_var: Find variable in this set, return its offset (-1 if not found).
//-----------------------------------------------------------------------------

int rvars_c::find_var(var_t var) const
{
	rvar_t p;

	// Check consistency:
	CHECK(valid(), "rvars_c::find_var: Bad object");

	// Check argument:
	CHECK(var != 0, "rvars_c::find_var: var is NIL");
	CHECK(var->valid(), "rvars_c::find_var: var is invalid");

	// Linear search (rules seldom have many variables):
	for(p = First; p; p = p->get_next())
		if(p->get_var() == var)
			return(p->get_offset());

	// Not found:
	return(-1);
}

//-----------------------------------------------------------------------------
// add_var: Add a variable to this current set, return its offset.
//-----------------------------------------------------------------------------

int rvars_c::add_var(var_t var)
{
	rvar_t p;

	// Check consistency:
	CHECK(valid(), "rvars_c::add_var: Bad object");

	// Check argument:
	CHECK(var != 0, "rvars_c::add_var: var is NIL");
	CHECK(var->valid(), "rvars_c::add_var: var is invalid");

	// Check limit for number of variables:
	CHECK(NumVars < VER_MAX_VARS, "rvars_c::add_var: too many vars");

	// Check that the variable is not already contained in the set:
	CHECK(find_var(var) == -1, "rvars_c::add_var: Duplicate variable");

	// Create new list element and link it to the list (at the end):
	p = new rvar_c(var, NumVars);
	if(First == 0)
		First = p;
	else
		Last->set_next(p);
	Last = p;
	return(NumVars++);
}

//-----------------------------------------------------------------------------
// get_var: Return variable with given offset.
//-----------------------------------------------------------------------------

var_t rvars_c::get_var(int offset) const
{
	rvar_t p;
	int i;

	// Check consistency:
	CHECK(valid(), "rvars_c::get_var: Bad object");

	// Check argument:
	CHECK(offset >= 0, "rvars_c::get_var: offset is negative");
	CHECK(offset < NumVars, "rvars_c::get_var: offset is too large");

	// Linear search (rules seldom have many variables):
	p = First;
	for(i = offset; i-- > 0; ) {
		CHECK(p != 0, "rvars_c::get_var: Variable list too short");
		p = p->get_next();
	}

	// Return variable:
	return(p->get_var());
}
